import Foundation
import Combine

enum SRVResolverAction: ModuleAction {
    var moduleName: ModuleName { "Resolver" }

    case startResolve(_ forDomain: String)
    case recordsFound(_ records: [SRVRecord])
    case resolvingError(_ error: SRVResolverError?)
}

struct SRVResolverState: ModuleState {
    var moduleName: ModuleName { "Resolver" }
    var foundedRecords: [SRVRecord]
}

struct SRVResolverModule: Module {
    var moduleName: ModuleName { "Resolver" }

    func getInitState() -> OptionalState {
        SRVResolverState.init(foundedRecords: [])
    }

    func reduce(state: inout any ModuleState, action: any ModuleAction) {
        guard let action = action as? SRVResolverAction, var newState = state as? SRVResolverState else {
            return
        }
        switch action {
        case .recordsFound(let records):
            newState.foundedRecords = records

        default:
            break
        }
        state = newState
    }

    func middleware(state: GlobalState, action: any ModuleAction) -> AnyPublisher<any ModuleAction, Never> {
        switch action {
        case let act as ClientAction:
            switch act {
            case .reconnect(let jid):
                return Just(SRVResolverAction.startResolve(jid.domainPart))
                    .eraseToAnyPublisher()
            }

        case let act as SRVResolverAction:
            switch act {
            case .startResolve(let domain):
                return Future<any ModuleAction, Never> { promise in
                    Task {
                        do {
                            let records = try await SRVResolver.resolve(domain: domain)
                            promise(.success(SRVResolverAction.recordsFound(records)))
                        } catch {
                            promise(.success(SRVResolverAction.resolvingError(error as? SRVResolverError)))
                        }
                    }
                }
                    .eraseToAnyPublisher()

            default:
                return Empty().eraseToAnyPublisher()
            }

        default:
            return Empty().eraseToAnyPublisher()
        }
    }
}
