/*
In 99,99% of time YOU DON'T NEEDED TO CHANGE ANYTHING in this file!

This file declare main state object for whole XMPP client
and reducers/actions/middleware types. Core of XMPP client.
*/

import Foundation
import Combine

enum StoreErrors: String, Error {
    case moduleWithSameNameExists = "Module with same name already exist. Try to use 'replaceModule(_ module: any Module)' instead"
    case moduleDoesntExists = "Trying replace module which not exist. Try 'addModule(_ module: any Module)' instead"
    case onlyStructStateAllowed = "Some of modules defines ModuleState as a class. Only struct here allowed"
}

public final class Store: ObservableObject {
    private let serialQueue = DispatchQueue(label: "im.narayana.snikket.xmppclient.serial.queue", qos: .userInteractive)
    private var modules: [any Module] = []
    private var middlewareCancellables: Set<AnyCancellable> = []
    private(set) var state: GlobalState = [:]

    init() {
        registerDefaultModules()
    }

    // Run reducers/middlewares
    func dispatch(_ action: any ModuleAction) {
        serialQueue.sync { [weak self] in
            guard let wSelf = self else {
                return
            }
            let newState = wSelf.dispatch(wSelf.state, action)
            wSelf.state = newState
        }
    }

    private func dispatch(_ currentState: GlobalState, _ action: any ModuleAction) -> GlobalState {
        let startTime = CFAbsoluteTimeGetCurrent()
        var newState =  currentState

        // Do reducing
        // Module can reduce only related state
        let name = action.moduleName
        if let module = modules.first(where: { $0.moduleName == name }), var moduleState = newState[name] {
            // make changes
            module.reduce(state: &moduleState, action: action)

            // Append changes to global state
            newState[moduleState.moduleName] = moduleState
        }

        // Dispatch all middleware functions
        // Middleware of any module can handle any action
        for module in modules {
            module.middleware(state: newState, action: action)
                .receive(on: DispatchQueue.main)
                .sink(receiveValue: dispatch)
                .store(in: &middlewareCancellables)
        }

        // Check performance
        let timeElapsed = CFAbsoluteTimeGetCurrent() - startTime
        if timeElapsed > 0.05 {
            print(
                """
                --
                (Ignore this warning ONLY in case, when execution is paused by your breakpoint)
                🕐Execution time: \(timeElapsed)
                ❌WARNING! Some reducers/middlewares work too long! It will lead to issues in production build!
                Because of execution each action is synchronous the any stuck will reduce performance dramatically.
                Probably you need check which part of reducer/middleware should be async (wrapped with Futures, as example)
                --
                """
            )
        }

        //
        return newState
    }
}

public extension Store {
    func addModule(_ module: any Module) throws {
        // check that module is unique
        if modules.map({ $0.moduleName }).contains(module.moduleName) {
            throw StoreErrors.moduleWithSameNameExists
        }

        // check that module state is struct
        if let moduleState = module.getInitState() {
            if Mirror(reflecting: moduleState).displayStyle == .class {
                throw StoreErrors.onlyStructStateAllowed
            }
            state[module.moduleName] = moduleState
        }

        // apply async dispatch for modules which should react on some outside events
        // force casting requires here
        // swiftlint:disable force_cast
        if var module = module as? ModuleContinuousActing {
            module.asyncDispatch = self.dispatch
            modules.append(module as! any Module)
        } else {
            modules.append(module)
        }
    }

    func replaceModule(_ module: any Module) throws {
        if !modules.map({ $0.moduleName }).contains(module.moduleName) {
            throw StoreErrors.moduleDoesntExists
        }
        modules.removeAll(where: { $0.moduleName == module.moduleName })
        state.removeValue(forKey: module.moduleName)
        try addModule(module)
    }
}

private extension Store {
    func registerDefaultModules() {
        do {
            try addModule(Logger())
            try addModule(SRVResolverModule())
            try addModule(SocketModule())
        } catch let err as StoreErrors {
            fatalError(err.rawValue)
        } catch {
            fatalError("\(error)")
        }
    }
}
